package explicit;

import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.PrimitiveIterator;
import java.util.Spliterator;
import java.util.Spliterators;
import java.util.stream.IntStream;
import java.util.stream.StreamSupport;

import common.iterable.FilteringIterator;
import common.iterable.SingletonIterator;

public abstract class SuccessorsIterator implements PrimitiveIterator.OfInt
{
	public abstract boolean successorsAreDistinct();
	public abstract boolean hasNext();
	public abstract int nextInt();

	public SuccessorsIterator distinct()
	{
		if (successorsAreDistinct()) {
			return this;
		} else {
			return new SuccessorsIteratorFromOfInt(FilteringIterator.dedupe(this), true);
		}
	}

	public IntStream stream()
	{
		return StreamSupport.intStream(
				Spliterators.spliteratorUnknownSize(this, successorsAreDistinct() ? Spliterator.DISTINCT : 0),
				false);
	}

	private static class SuccessorsIteratorFromOfInt extends SuccessorsIterator {
		private java.util.PrimitiveIterator.OfInt it;
		private boolean distinct;

		public SuccessorsIteratorFromOfInt(PrimitiveIterator.OfInt it, boolean distinct)
		{
			this.it = it;
			this.distinct = distinct;
		}

		@Override
		public boolean successorsAreDistinct()
		{
			return distinct;
		}

		@Override
		public boolean hasNext()
		{
			return it.hasNext();
		}

		@Override
		public int nextInt()
		{
			return it.nextInt();
		}
	}

	private static class SuccessorsIteratorFromIterator extends SuccessorsIterator {
		private Iterator<Integer> it;
		private boolean distinct;

		public SuccessorsIteratorFromIterator(Iterator<Integer> it, boolean distinct)
		{
			this.it = it;
			this.distinct = distinct;
		}

		@Override
		public boolean hasNext()
		{
			return it.hasNext();
		}

		@Override
		public Integer next()
		{
			return it.next();
		}

		@Override
		public int nextInt()
		{
			return it.next();
		}

		@Override
		public boolean successorsAreDistinct()
		{
			return distinct;
		}
	};

	private static class SuccessorsIteratorEmpty extends SuccessorsIterator {
		@Override
		public boolean hasNext()
		{
			return false;
		}

		@Override
		public int nextInt()
		{
			throw new NoSuchElementException();
		}

		@Override
		public boolean successorsAreDistinct()
		{
			return true;
		}
	};

	private static class ChainedSuccessorsIterator extends SuccessorsIterator {
		private Iterator<SuccessorsIterator> iterators;
		private SuccessorsIterator current;
		private boolean distinct;

		public ChainedSuccessorsIterator(Iterator<SuccessorsIterator> iterators)
		{
			this.iterators = iterators;
			current = iterators.hasNext() ? iterators.next() : null;
			if (current != null && !iterators.hasNext()) {
				// only a single successor iterator, can inherit elementsAreDistinct
				distinct = current.successorsAreDistinct();
			} else {
				// can not guarantee that successors are distinct
				distinct = false;
			}
		}

		@Override
		public boolean hasNext()
		{
			if (current == null) {
				return false;
			}

			if (current.hasNext()) {
				// the current iterator has another element
				return true;
			}

			// the current iterator has no more elements,
			// search for the next iterator that as an element
			while (iterators.hasNext()) {
				// consider the next iterator
				current = iterators.next();
				if (current.hasNext()) {
					// iterator has element, keep current and return true
					return true;
				}
			}

			// there are no more iterators / elements
			current = null;
			return false;
		}

		@Override
		public int nextInt()
		{
			if (!hasNext()) {
				throw new NoSuchElementException();
			}
			return current.nextInt();
		}

		@Override
		public boolean successorsAreDistinct()
		{
			return distinct;
		}
	};

	public static SuccessorsIterator from(Iterator<Integer> it, boolean distinctElements)
	{
		return new SuccessorsIteratorFromIterator(it, distinctElements);
	}

	public static SuccessorsIterator from(PrimitiveIterator.OfInt it, boolean distinctElements)
	{
		return new SuccessorsIteratorFromOfInt(it, distinctElements);
	}

	public static SuccessorsIterator fromSingleton(int i)
	{
		return new SuccessorsIteratorFromOfInt(new SingletonIterator.OfInt(i), true);
	}

	public static SuccessorsIterator empty()
	{
		return new SuccessorsIteratorEmpty();
	}

	public static SuccessorsIterator chain(Iterator<SuccessorsIterator> iterators)
	{
		return new ChainedSuccessorsIterator(iterators);
	}
}
